class Updater:
    @staticmethod
    def merge_mssql(source_schema: str, source_table: str, target_schema: str, target_table: str, columns: list[str],
                    id_column: str, delete_unmatched: bool = True):
        stage = f'{source_schema}.{source_table}'
        id_column = f'[{id_column}]'
        location = f'{target_schema}.{target_table}'
        clean_columns = [f'[{column}]' for column in columns]
        source_columns = ', '.join([f'b.{column}' for column in clean_columns])
        column_string = ', '.join(clean_columns)
        update_check = ' or '.join(
            [f'a.{column} <> b.{column} or (a.{column} is null and b.{column} is not null) ' for column in
             clean_columns if column != id_column]
        )
        update_check = f' and ({update_check})'
        update_columns = ', '.join([f'a.{column} = b.{column}' for column in clean_columns if column != id_column])
        query = (
            f'merge {location} a using {stage} b on a.{id_column} = b.{id_column} '
            f'when matched {update_check} then update set {update_columns} '
            f'when not matched by target then insert ({column_string}) values ({source_columns})'
        )
        if delete_unmatched:
            query = f'{query} when not matched by source then delete'
        return query

    @staticmethod
    def upsert_mssql(source_schema: str, source_table: str, target_schema: str, target_table: str, columns: list[str],
                     id_column: str):
        stage = f'{source_schema}.{source_table}'
        location = f'{target_schema}.{target_table}'
        clean_columns = [f'[{column}]' for column in columns]
        column_string = ', '.join(clean_columns)
        stage_columns = [f's.{column}' for column in clean_columns]
        stage_column_string = ', '.join(stage_columns)
        delete_dupes_query = (
            f'Delete from {stage} from {stage} s where exists (select '
            f'{stage_column_string} intersect select {column_string} from {location})'
        )
        delete_old_query = (
            f'delete from {location} where {id_column} in ( '
            f'select {id_column} from {stage} intersect select {id_column} from {location})'
        )
        insert_query = (
            f'insert into {location} ({column_string}) select {column_string} from {stage}'
        )
        query = f'{delete_dupes_query}; {delete_old_query}; {insert_query};'
        return query

    @staticmethod
    def append_mssql(source_schema: str, source_table: str, target_schema: str, target_table: str, columns: list[str]):
        stage = f'{source_schema}.{source_table}'
        location = f'{target_schema}.{target_table}'
        clean_columns = [f'[{column}]' for column in columns]
        column_string = ','.join(clean_columns)
        query = (
            f'insert into {location} ({column_string}) select {column_string} from {stage}'
            f'except select {column_string} from {location}'
        )
        return query



import unittest
import pandas as pd
from datetime import datetime
from src.etl.query.creator import Creator


class TestTableMaker(unittest.TestCase):

    def test_make_mssql_table(self):
        df = pd.DataFrame({
            'id_column': [2147483649, 2147483650, 2147483651],
            'int_column': [1, 2, 3],
            'float_column': [1.1, 2.2, 3.3],
            'date_column': ['2021-01-01', '2021-01-02', '2021-01-03'],
            'bool_column': [True, False, True],
            'str_column': ['a', 'bb', 'ccc'],
            'empty_column': [None, None, None]
        })

        expected_query = (
            "create table dbo.[test_table] ("
            "[id_column] bigint constraint pk_test_table_id_column primary key, "
            "[int_column] tinyint, "
            "[float_column] decimal(10, 2), "
            "[date_column] datetime2, "
            "[bool_column] bit, "
            "[str_column] nvarchar(23), "
            "[empty_column] nvarchar(max), "
            "system_record_start datetime2 generated always as row start "
            "constraint df_test_table_system_record_start default sysutcdatetime() not null, "
            "system_record_end datetime2 generated always as row end "
            "constraint df_test_table_system_record_end default sysutcdataetime() not null, "
            "period for system_time(system_record_start, system_record_end)) "
            "with (system_versioning = on (history_table = dbo.[test_table_history]));"
        )

        actual_query = Creator.create_mssql_table(df, 'dbo', 'test_table',
                                                  primary_key='id_column', history=True)

        self.assertEqual(expected_query, actual_query)


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
from unittest.mock import Mock, patch
from src.etl.database.loader import Loader


class TestLoader(unittest.TestCase):

    @patch('src.etl.database.loader.insert_to_db')
    def test_insert_to_mssql_table(self, mock_insert):
        cursor = Mock()
        df = pd.DataFrame({
            'id': [1, 2, 3],
            'name': ['Alice', 'Bob', 'Charlie'],
            'value': [10.5, 20.3, 30.7]
        })
        schema = 'dbo'
        table = 'test_table'

        Loader.insert_to_mssql_table(cursor, df, schema, table)

        self.assertTrue(mock_insert.called)
        self.assertGreater(mock_insert.call_count, 0)


if __name__ == '__main__':
    unittest.main()



import urllib
from sqlalchemy import create_engine


class Connector:
    @staticmethod
    def get_mssql_trusted_connection(host: str, instance: str, database: str):
        connection_string = f"SERVER={host}\\{instance};DATABASE={database};Trusted_Connection=yes;"
        quoted = urllib.parse.quote_plus("DRIVER={ODBC DRIVER 17 for SQL SERVER};" + connection_string)
        engine = create_engine('mssql+pyodbc:////?odbc_connect={}'.format(quoted))
        return engine.connect().connection

    @staticmethod
    def get_mssql_user_connection(host: str, instance: str, database: str, username: str, password: str):
        connection_string = f"SERVER={host}\\{instance};DATABASE={database};UID={username};PWD={password};"
        quoted = urllib.parse.quote_plus("DRIVER={ODBC DRIVER 17 for SQL SERVER};" + connection_string)
        engine = create_engine('mssql+pyodbc:////?odbc_connect={}'.format(quoted))
        return engine.connect().connection

    @staticmethod
    def get_postgres_user_connection(host: str, port: int, database: str, username: str, password: str):
        connection_string = f'postgresql://{username}:{password}@{host}:{port}/{database}'
        engine = create_engine(connection_string)
        return engine.connect().connection

    @staticmethod
    def get_mysql_user_connection(host: str, database: str, username: str, password: str):
        connection_string = f'mysql+pymysql://{username}:{password}@{host}/{database}?charset=utf8mb4'
        engine = create_engine(connection_string)
        return engine.connect().connection




from .. import constants
import numpy as np
import pandas as pd
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, MofNCompleteColumn
from rich import print


def insert_to_db(column_string, cursor, data_list, location, values):
    value_list = " union ".join(['select {}'.format(value) for value in values])
    execute_query = (
        f"insert into {location} ({column_string}) {value_list}"
    )
    try:
        cursor.execute(execute_query, data_list)
    except Exception as e:
        print(execute_query)
        print(data_list)
        raise e


class Loader:
    @staticmethod
    def insert_to_mssql_table(cursor, df: pd.DataFrame, schema: str, table: str):
        column_list = df.columns.tolist()
        column_list = [f'[{column}]' for column in column_list]
        column_string = ", ".join(column_list)
        location = f"{schema}.[{table}]"
        row_values = []
        for column in df.columns:
            series = df[column]
            series_type = series.dtype
            str_column = series.apply(str)
            max_size = str_column.str.len().max()
            if max_size > 256:
                row_values.append('cast ( ? as nvarchar(max))')
            else:
                row_values.append('?')
            # switches from numpy class to python class for bool float and int
            if series_type in constants.NUMPY_BOOL_TYPES or series_type in constants.NUMPY_INT_TYPES or series_type in constants.NUMPY_FLOAT_TYPES:
                df[column] = series.tolist()
        Loader.insert_to_table(column_string, cursor, df, location, row_values, table)

    @staticmethod
    def insert_to_mysql_table(cursor, df: pd.DataFrame, schema: str, table: str):
        column_list = df.columns.tolist()
        column_list = [f'`{column}`' for column in column_list]
        column_string = ", ".join(column_list)
        location = f'{schema}.`{table}`'
        row_values = []
        for column in df.columns:
            series = df[column]
            series_type = series.dtype
            str_column = series.apply(str)
            max_size = str_column.str.len().max()
            if max_size > 255:
                row_values.append('cast ( %s as varchar(21844))')
            else:
                row_values.append('%s')
            # switches from numpy class to python class for bool float and int
            if series_type in constants.NUMPY_BOOL_TYPES or series_type in constants.NUMPY_INT_TYPES or series_type in constants.NUMPY_FLOAT_TYPES:
                df[column] = series.tolist()
        Loader.insert_to_table(column_string, cursor, df, location, row_values, table)

    @staticmethod
    def insert_to_table(column_string, cursor, df, location, row_values, table):
        row_value_list = ", ".join(row_values)
        df = df.replace({np.nan: None})
        with Progress(TextColumn("[progress.description]{task.description}"), BarColumn(), TaskProgressColumn(),
                      MofNCompleteColumn()) as progress:
            total = df.shape[0]
            values = []
            data_list = []
            data_count = 0
            row_count = 0
            upload_task = progress.add_task(f'loading {table}', total=total)
            for row in df.itertuples(index=False, name=None):
                row_size = len(row)
                row_count += 1
                data_count += row_size
                values.append(row_value_list)

                data_list.extend(row)
                next_size = data_count + row_size
                if next_size >= 2000:
                    insert_to_db(column_string, cursor, data_list, location, values)
                    progress.update(upload_task, advance=row_count)
                    values = []
                    data_list = []
                    data_count = 0
                    row_count = 0
            if row_count > 0:
                insert_to_db(column_string, cursor, data_list, location, values)
                progress.update(upload_task, advance=row_count)



import hashlib
import unittest
import pandas as pd
from src.etl.dataframe.cleaner import compute_hash, Cleaner


class TestCleanFunctions(unittest.TestCase):

    def test_compute_hash(self):
        self.assertEqual(compute_hash('test'), hashlib.sha1('test'.encode()).hexdigest())


class TestCleaner(unittest.TestCase):

    def setUp(self):
        self.df = pd.DataFrame({
            'MixedCase': ['value1', 'value2'],
            'with spaces': ['value3', 'value4'],
            'with$Special#Chars': ['value5', 'value6']
        })

    def test_column_names_to_snake_case(self):
        expected_columns = ['mixed_case', 'with_spaces', 'with_dollars_special_num_chars']
        Cleaner.column_names_to_snake_case(self.df)
        self.assertEqual(expected_columns, self.df.columns.tolist())

    def test_column_names_to_pascal_case(self):
        expected_columns = ['MixedCase', 'WithSpaces', 'WithDollarsSpecialNumChars']
        Cleaner.column_names_to_pascal_case(self.df)
        self.assertEqual(expected_columns, self.df.columns.tolist())

    def test_parse_numbers(self):
        df = pd.DataFrame({
            'ints': ['$1,000', '2,000', '3,000', '244385297.0'],
            'floats': ['$1,000.55', '2,000.66', '3,000.33', '1209347']
        })
        clean_df = Cleaner.clean_numbers(df)
        self.assertEqual(clean_df['ints'].dtype, 'int64')
        self.assertEqual(clean_df['floats'].dtype, 'float64')

    def test_clean_dates(self):
        df = pd.DataFrame({
            'dates': ['2021-01-01', '01/02/2021']
        })
        clean_df = Cleaner.clean_dates(df)
        self.assertEqual(clean_df['dates'].dtype, 'datetime64[ns]')

    def test_clean_bools(self):
        df = pd.DataFrame({
            'bools': ['yes', 'no', 'true', 'false']
        })
        clean_df = Cleaner.clean_bools(df)
        self.assertEqual(clean_df['bools'].dtype, 'bool')

    def test_clean_all_types(self):
        df = pd.DataFrame({
            'numbers': ['$1,000', '248166676', '3,000', '%100'],
            'dates': ['2021-01-01', '01/02/2021', '2021-01-01', '01/02/2021'],
            'bools': ['yes', 'no', 'true', 'false'],
            'empty': [None, None, None, None]
        })
        clean_df = Cleaner.clean_all_types(df)
        self.assertEqual(clean_df['numbers'].dtype, pd.Int64Dtype.name)
        self.assertEqual(clean_df['dates'].dtype, 'datetime64[ns]')
        self.assertEqual(clean_df['bools'].dtype, pd.BooleanDtype.name)

    def test_clean_df(self):
        df = pd.DataFrame({
            'numbers': [None, '248166676', '3,000', '%100'],
            'bools': [None, 'no', 'true', 'false'],
            'empty': [None, None, None, None]
        })
        clean_df = Cleaner.clean_df(df)
        self.assertEqual(clean_df.shape, (3, 2))


    def test_generate_hash_column(self):
        df = pd.DataFrame({
            'col1': ['value1', 'value2'],
            'col2': ['value3', 'value4']
        })
        expected_hash = df['col1'].apply(str) + df['col2'].apply(str)
        expected_hash = expected_hash.apply(compute_hash)
        expected_hash.name = 'hash'
        result_df = Cleaner.generate_hash_column(df, ['col1', 'col2'], 'hash')
        result_series = result_df['hash']
        pd.testing.assert_series_equal(result_series, expected_hash)

    def test_coalesce_columns(self):
        df = pd.DataFrame({
            'col1': [None, 'B'],
            'col2': ['A', None],
        })
        expected_df = pd.DataFrame({
            'col1': ['A', 'B']
        })
        result_df = Cleaner.coalesce_columns(df, ['col1', 'col2'], 'col1',drop=True)
        self.assertEqual(result_df.shape, (2, 1))
        pd.testing.assert_frame_equal(result_df, expected_df)


if __name__ == '__main__':
    unittest.main()



import importlib
import sys
import unittest
from unittest.mock import patch, Mock

sys.path.insert(0, '/home/jedi/Projects/Python/etl_utilities/src/etl/database')

# Import after updating sys.path
from connector import Connector


class TestConnector(unittest.TestCase):

    @patch('connector.urllib.parse.quote_plus')
    @patch('connector.create_engine')
    def test_get_mssql_trusted_connection(self, mock_create_engine, mock_quote_plus):
        mock_engine = Mock()
        mock_engine.connect.return_value.connection = 'connection-object'
        mock_create_engine.return_value = mock_engine
        host, instance, database = 'host', 'instance', 'database'
        Connector.get_mssql_trusted_connection(host, instance, database)
        mock_quote_plus.assert_called()
        mock_create_engine.assert_called()

    @patch('connector.urllib.parse.quote_plus')
    @patch('connector.create_engine')
    def test_get_mssql_user_connection(self, mock_create_engine, mock_quote_plus):
        mock_engine = Mock()
        mock_engine.connect.return_value.connection = 'connection-object'
        mock_create_engine.return_value = mock_engine
        Connector.get_mssql_user_connection('host', 'instance', 'database', 'username', 'password')
        mock_quote_plus.assert_called()
        mock_create_engine.assert_called()

    @patch('connector.create_engine')
    def test_get_postgres_user_connection(self, mock_create_engine):
        mock_engine = Mock()
        mock_engine.connect.return_value.connection = 'connection-object'
        mock_create_engine.return_value = mock_engine
        Connector.get_postgres_user_connection('host', 8080, 'database', 'username', 'password')
        mock_create_engine.assert_called()

    @patch('connector.create_engine')
    def test_get_mysql_user_connection(self, mock_create_engine):
        mock_engine = Mock()
        mock_engine.connect.return_value.connection = 'connection-object'
        mock_create_engine.return_value = mock_engine
        Connector.get_mysql_user_connection('host', 'database', 'username', 'password')
        mock_create_engine.assert_called()


if __name__ == '__main__':
    unittest.main()



import pandas as pd
from ..dataframe.analyzer import Analyzer
from rich import print


class Creator:
    @staticmethod
    def create_mssql_table(df: pd.DataFrame, schema: str, table: str, primary_key: str = None,
                           unique_columns: list[str] = None, history: bool = False,
                           varchar_padding: int = 20, float_precision: int = 10, decimal_places: int = 2,
                           generate_id: bool = False):
        location = f'{schema}.[{table}]'
        column_metadata = Analyzer.generate_column_metadata(df, primary_key, unique_columns, decimal_places)
        column_type_list = []
        if generate_id:
            id_string = f'id int identity constraint pk_{table}_id primary key'
            column_type_list.append(id_string)
        for column in column_metadata:
            column_string = None
            column_name = column['column_name']
            if column['is_empty']:
                print(f"{column_name} is empty - setting to nvarchar(max)")
                column_string = f'[{column_name}] nvarchar(max)'
                column_type_list.append(column_string)
                continue
            if column['data_type'] == 'datetime':
                column_string = f'[{column_name}] datetime2'
            if column['data_type'] == 'float':
                if float_precision < column['float_precision']:
                    float_precision = column['float_precision']
                column_string = f'[{column_name}] decimal({float_precision}, {decimal_places})'
            if column['data_type'] == 'integer':
                if column['smallest_num'] < -2147483648 or column['biggest_num'] > 2147483648:
                    column_string = f'[{column_name}] bigint'
                if column['smallest_num'] >= -2147483648 and column['biggest_num'] <= 2147483648:
                    column_string = f'[{column_name}] int'
                if column['smallest_num'] >= -32768 and column['biggest_num'] <= 32768:
                    column_string = f'[{column_name}] smallint'
                if column['smallest_num'] >= 0 and column['biggest_num'] <= 255:
                    column_string = f'[{column_name}] tinyint'
            if column['data_type'] == 'boolean':
                column_string = f'[{column_name}] bit'
            if column['data_type'] == 'string':
                padded_length = int(column['max_str_size'] + varchar_padding)
                if padded_length >= 4000:
                    column_string = f'[{column_name}] nvarchar(max)'
                else:
                    column_string = f'[{column_name}] nvarchar({padded_length})'
            if column['is_id']:
                column_string += f' constraint pk_{table}_{column_name} primary key'
            if column['is_unique']:
                column_string += f' constraint ak_{table}_{column_name} unique'
            column_type_list.append(column_string)

        column_type_string = ', '.join(column_type_list)
        create_query = f'create table {location} ({column_type_string});'
        if history:
            history_location = f'{schema}.[{table}_history]'
            history_insert = (
                f'{column_type_string}, '
                'system_record_start datetime2 generated always as row start '
                f'constraint df_{table}_system_record_start '
                'default sysutcdatetime() not null, '
                'system_record_end datetime2 generated always as row end '
                f'constraint df_{table}_system_record_end '
                'default sysutcdataetime() not null, '
                'period for system_time(system_record_start, system_record_end)'
            )
            create_query = (
                f'create table {location} ({history_insert}) with ('
                f'system_versioning = on (history_table = {history_location}));'
            )
        return create_query

    @staticmethod
    def create_mariadb_table(df: pd.DataFrame, schema: str, table: str, primary_key: str = None,
                           unique_columns: list[str] = None, history: bool = False,
                           varchar_padding: int = 20, float_precision: int = 10, decimal_places: int = 2,
                           generate_id: bool = False):
        location = f'{schema}.{table}'
        column_metadata = Analyzer.generate_column_metadata(df, primary_key, unique_columns, decimal_places)
        column_type_list = []
        if generate_id:
            id_string = f'id int auto_increment, constraint pk_{table}_id primary key (id)'
            column_type_list.append(id_string)
        for column in column_metadata:
            column_string = None
            column_name = column['column_name']
            if column['is_empty']:
                print(f"{column_name} is empty - skipping")
                continue
            if column['data_type'] == 'datetime':
                column_string = f'`{column_name}` datetime'
            if column['data_type'] == 'float':
                if float_precision < column['float_precision']:
                    float_precision = column['float_precision']
                column_string = f'`{column_name}` decimal({float_precision}, {decimal_places})'
            if column['data_type'] == 'integer':
                if column['smallest_num'] < -2147483648 or column['biggest_num'] > 2147483648:
                    column_string = f'`{column_name}` bigint'
                if column['smallest_num'] >= -2147483648 and column['biggest_num'] <= 2147483648:
                    column_string = f'`{column_name}` int'
                if column['smallest_num'] >= -32768 and column['biggest_num'] <= 32768:
                    column_string = f'`{column_name}` smallint'
                if column['smallest_num'] >= -128 and column['biggest_num'] <= 127:
                    column_string = f'`{column_name}` tinyint'
            if column['data_type'] == 'boolean':
                column_string = f'`{column_name}` bit'
            if column['data_type'] == 'string':
                padded_length = int(column['max_str_size'] + varchar_padding)
                if padded_length > 21844:
                    print(f"{column_name} has data too large for storing - skipping")
                    continue
                else:
                    column_string = f'`{column_name}` varchar({padded_length})'
            if column['is_id']:
                key_string = f'constraint pk_{table}_{column_name} primary key ({column_name})'
                column_type_list.append(key_string)
            if column['is_unique']:
                key_string = f'constraint ak_{table}_{column_name} unique ({column_name})'
                column_type_list.append(key_string)
            column_type_list.append(column_string)

        column_type_string = ', '.join(column_type_list)
        create_query = f'create table {location} ({column_type_string});'
        if history:
            create_query = (
                f'create table {location} ({column_type_string}) with ('
                f'system_versioning;'
            )
        return create_query


import numpy as np
import pandas as pd

DB_INT_TYPES = ['bigint', 'int', 'smallint', 'tinyint', 'mediumint', 'integer', 'bit']
DB_FLOAT_TYPES = ['decimal', 'numeric', 'float', 'double']
DB_STR_TYPES = ['varchar', 'nvarchar', 'char', 'nchar']
DB_BOOL_TYPES = ['bit', 'tinyint', 'bool', 'boolean']
DB_DATE_TYPES = ['date', 'datetime', 'datetime2', 'time', 'timestamp', 'smalldatetime', 'year']
NUMPY_INT_TYPES = [np.int_, np.int64, np.int32, np.int8, 'Int64']
NUMPY_FLOAT_TYPES = [np.float64, np.float32, np.float16, 'Float64']
NUMPY_STR_TYPES = [np.str_, np.object_, 'string']
NUMPY_BOOL_TYPES = [np.bool_, np.True_, np.False_, pd.BooleanDtype, 'boolean']
NUMPY_DATE_TYPES = [np.datetime64, 'datetime64[ns]']

TYPE_MAPPINGS = {
    tuple(NUMPY_INT_TYPES): DB_INT_TYPES,
    tuple(NUMPY_FLOAT_TYPES): DB_FLOAT_TYPES,
    tuple(NUMPY_DATE_TYPES): DB_DATE_TYPES,
    tuple(NUMPY_STR_TYPES): DB_STR_TYPES,
    tuple(NUMPY_BOOL_TYPES): DB_BOOL_TYPES
}

